/* ANTINAT
 * =======
 * This software is Copyright (c) 2003 Malcolm Smith.
 * No warranty is provided, including but not limited to
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * This code is licenced subject to the GNU General
 * Public Licence (GPL).  See the COPYING file for more.
 */

#define _REENTRANT
#include "../../config.h"
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#include <DSNetwork.h>
#include <DSMutex.h>
#include <DSTSHashTable.h>
#include <DSConfigFile.h>
#include <DSThread.h>
#include "../Connection.h"
#include "../loggers.h"
#include "../Locks.h"
#include "../Shared.h"

struct tm lastlog;

DSMutex filelock;
DSMutex lastlog_lock;

DSTSHashTable *logAddrMonth;
DSTSHashTable *logAddrDay;
DSTSHashTable *logAddrHour;
DSTSHashTable *logAddrMinute;

BOOL useAddrMonth;
BOOL useAddrDay;
BOOL useAddrHour;
BOOL useAddrMinute;

DSTSHashTable *logUserMonth;
DSTSHashTable *logUserDay;
DSTSHashTable *logUserHour;
DSTSHashTable *logUserMinute;

BOOL useUserMonth;
BOOL useUserDay;
BOOL useUserHour;
BOOL useUserMinute;

typedef struct stLogInfo
{
	unsigned long uploadBytes;
	unsigned long downloadBytes;
} stLogInfo;

typedef struct stLogPath
{
	DSTSHashTable * src;
	char path[256];
} stLogPath;

void * outputLogChildThread(void * pdata)
{
	DSList * lst;
	DSListElement * le;
	stLogInfo * li;
	struct stLogPath * param;
	FILE * fp;
	param=(struct stLogPath *)pdata;
	param->src->lock();
#ifdef WITH_DEBUG
	printf("Attempting to update summary log %s\n",param->path);
#endif
	fp=fopen(param->path,"a");
	if (fp) {
		fprintf(fp,"\n");
		lst=param->src->getList();
#if LIBDS_VERSION >= 0x00010300
		lst->sortAsString();
#endif
		for (le=lst->getHead();le!=NULL;le=le->getNext()) {
			li=(stLogInfo *)le->getDataPtr();
			fprintf(fp,"%s %lu %lu\n",le->getKeyString(),li->downloadBytes,li->uploadBytes);
		}
		delete lst;
		fclose(fp);
	}
	param->src->close();
	param->src->unlock();
	delete param->src;
	free(param);
	return NULL;
}



BOOL outputLog(DSTSHashTable ** src, char * path)
{
	struct stLogPath * logpath;
	DSThread thr;
	logpath=(struct stLogPath *)malloc(sizeof(stLogPath));
	logpath->src=*src;
	strncpy(logpath->path,path,sizeof(logpath->path));
	*src=new DSTSHashTable();
	thr.exec(outputLogChildThread, logpath);
	thr.detach();
	return TRUE;
}

BOOL checkForLogOutput()
{
	BOOL monthChanged;
	BOOL dayChanged;
	BOOL hourChanged;
	BOOL minuteChanged;
	time_t now;
	struct tm * tm;
#ifdef HAVE_LOCALTIME_R
	struct tm realtm;
#endif

#ifdef WITH_DEBUG
	printf("Attempting to find if log output necessary\n");
#endif

	monthChanged=FALSE;
	dayChanged=FALSE;
	hourChanged=FALSE;
	minuteChanged=FALSE;
	time(&now);

	lastlog_lock.lock();
#ifdef HAVE_LOCALTIME_R
	tm=localtime_r(&now, &realtm);
#else
	localtime_lock.lock();
	tm=localtime(&now);
#endif
	if ((tm->tm_year!=lastlog.tm_year)||(tm->tm_mon!=lastlog.tm_mon))
		monthChanged=TRUE;
	if ((tm->tm_mday!=lastlog.tm_mday)||(monthChanged))
		dayChanged=TRUE;
	if ((tm->tm_hour!=lastlog.tm_hour)||(dayChanged))
		hourChanged=TRUE;
	if ((tm->tm_min!=lastlog.tm_min)||(hourChanged))
		minuteChanged=TRUE;

	memcpy(&lastlog,tm,sizeof(lastlog));

#ifndef HAVE_LOCALTIME_R
	localtime_lock.unlock();
#endif

	if (monthChanged) {
		if (useAddrMonth)
			outputLog(&logAddrMonth, logdir "/addr_month.log");
		if (useUserMonth)
			outputLog(&logUserMonth, logdir "/user_month.log");
	}

	if (dayChanged) {
		if (useAddrDay)
			outputLog(&logAddrDay, logdir "/addr_day.log");
		if (useUserDay)
			outputLog(&logUserDay, logdir "/user_day.log");
	}

	if (hourChanged) {
		if (useAddrHour)
			outputLog(&logAddrHour, logdir "/addr_hour.log");
		if (useUserHour)
			outputLog(&logUserHour, logdir "/user_hour.log");
	}

	if (minuteChanged) {
		if (useAddrMinute)
			outputLog(&logAddrMinute, logdir "/addr_min.log");
		if (useUserMinute)
			outputLog(&logUserMinute, logdir "/user_min.log");
	}
	lastlog_lock.unlock();
	return TRUE;
}

BOOL single_log(DSTSHashTable * dest, char * key, unsigned long down, unsigned long up)
{
	stLogInfo * li;
	dest->lock();
	li=(stLogInfo *)dest->getPtrValue(key);
	if (li) {
		li->uploadBytes+=up;
		li->downloadBytes+=down;
	} else {
		char * keycopy;
		li=(stLogInfo *)malloc(sizeof(stLogInfo));
		keycopy=(char *)malloc(strlen(key)+1);
		strcpy(keycopy,key);
		li->uploadBytes=up;
		li->downloadBytes=down;
		dest->insert(keycopy,li,DSListElement::CLEANUP_KEY_FREE|
				DSListElement::CLEANUP_VALUE_FREE);
	}

	dest->unlock();
	return TRUE;
}

BOOL cpp_real_log(Connection * conn)
{
	unsigned long upload;
	unsigned long download;
	char * logkey;

	checkForLogOutput();

	upload=conn->getUploadBytes();
	download=conn->getDownloadBytes();
	logkey=conn->getSourceAddrString();

	if (useAddrMonth)
		single_log(logAddrMonth, logkey, download, upload);

	if (useAddrDay)
		single_log(logAddrDay, logkey, download, upload);

	if (useAddrHour)
		single_log(logAddrHour, logkey, download, upload);
		
	if (useAddrMinute)
		single_log(logAddrMinute, logkey, download, upload);

	free(logkey);

	logkey=conn->getUser();
	if (logkey==NULL) logkey="Anonymous";

	if (useUserMonth)
		single_log(logUserMonth, logkey, download, upload);

	if (useUserDay)
		single_log(logUserDay, logkey, download, upload);

	if (useUserHour)
		single_log(logUserHour, logkey, download, upload);
		
	if (useUserMinute)
		single_log(logUserMinute, logkey, download, upload);

	return TRUE;
}

BOOL cpp_log(Connection * conn,int event)
{
	time_t now;
	struct tm * tm;
#ifdef HAVE_LOCALTIME_R
	struct tm realtm;
#endif
	switch(event) {
		case LOG_EVT_SERVERSTART:

			logAddrMonth=new DSTSHashTable();
			logAddrDay=new DSTSHashTable();
			logAddrHour=new DSTSHashTable();
			logAddrMinute=new DSTSHashTable();
			logUserMonth=new DSTSHashTable();
			logUserDay=new DSTSHashTable();
			logUserHour=new DSTSHashTable();
			logUserMinute=new DSTSHashTable();

			useAddrMonth=config.getValueInt("summary address log","month",0);
			useAddrDay=config.getValueInt("summary address log","day",1);
			useAddrHour=config.getValueInt("summary address log","hour",0);
			useAddrMinute=config.getValueInt("summary address log","minute",0);
			useUserMonth=config.getValueInt("summary user log","month",0);
			useUserDay=config.getValueInt("summary user log","day",1);
			useUserHour=config.getValueInt("summary user log","hour",0);
			useUserMinute=config.getValueInt("summary user log","minute",0);
			time(&now);
			lastlog_lock.lock();
#ifndef HAVE_LOCALTIME_R
			localtime_lock.lock();
			tm=localtime(&now);
#else
			tm=localtime_r(&now,&realtm);
#endif
			memcpy(&lastlog,tm,sizeof(lastlog));
#ifndef HAVE_LOCALTIME_R
			localtime_lock.unlock();
#endif
			lastlog_lock.unlock();
		break;
		case LOG_EVT_SERVERCLOSE:
			now=(time_t)-1;
			lastlog_lock.lock();
#ifndef HAVE_LOCALTIME_R
			localtime_lock.lock();
			tm=localtime(&now);
#else
			tm=localtime_r(&now,&realtm);
#endif
			memcpy(&lastlog,tm,sizeof(lastlog));
#ifndef HAVE_LOCALTIME_R
			localtime_lock.unlock();
#endif
			lastlog_lock.unlock();
			checkForLogOutput();
			
		break;
		case LOG_EVT_LOG:
		return cpp_real_log(conn);
		break;
	}
	return TRUE;
}

extern "C" {
int logger_type()
{
	return LOG_TYPE_CONNECTIONCLOSE;
}

BOOL log(void * conn,int event)
{
	return cpp_log((Connection *)conn,event);
}
}
