/* ANTINAT
 * =======
 * This software is Copyright (c) 2003 Malcolm Smith.
 * No warranty is provided, including but not limited to
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * This code is licenced subject to the GNU General
 * Public Licence (GPL).  See the COPYING file for more.
 */

#include "../config.h"
#include <DSThread.h>
#include "Connection.h"
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <DSNetwork.h>
#include <DSComponent.h>
#include <DSConfigFile.h>
#include <DSList.h>
#include <DSParams.h>
#include "loggers.h"
#include "moddecl.h"


DSNetwork server;
DSHashTable versions;
DSList filters;
DSList loggers;
DSHashTable addresses;
DSHashTable resolvers;
DSHashTable revresolve;
DSConfigFile config;

/*
This big messy function processes each incoming request.
*/
BOOL
HandleRequest ()
{
	DSNetwork * net;
	Connection * conn;
	DSThread thr;

	/* Accept an incoming request, or wait till one arrives. */
	net=server.acceptIncoming();
	if (!net) return FALSE;
	net->linger();
	conn=new Connection(net);

	thr.exec(ChildThread, conn);
	thr.detach();

	return TRUE;

}

BOOL logEvent(int event)
{
	DSListElement * le;
	proc_logger logger;
	for (le=loggers.getHead();le!=NULL;le=le->getNext()) {
		logger=(proc_logger)le->getDataPtr();
		logger(NULL,event);
	}
	return TRUE;
}

void daemonize()
{
	int fh;
	if (fork()!=0) exit(EXIT_SUCCESS);
	fh=open("/dev/null",O_RDWR,0666);
	close(0);
	close(1);
	close(2);
	dup2(fh, 0);
	dup2(fh, 1);
	dup2(fh, 2);
#ifdef HAVE_SETSID
	setsid();
#endif
}


/*
UNIX trivia - when is a problem a problem?  When you don't ignore it.
If you do nothing, well, you're not being ignorant enough.  You have
to be explicitly ignorant.
*/
void
ignorer (int x)
{
	signal (x, ignorer);
}

/*
 * TODO:
 * This is a potential deadlock condition.  If the thread handling
 * this signal holds the authdetails lock, it will attempt to take it
 * twice, and hold it forever; other threads will eventually try for
 * it and get caught in the traffic jam.
 *
 * I'm not sure what to do about it, tho. This signal can't be left to
 * one thread only because some OSes (thanks Linux!) send signals to
 * specific threads...
 *
 * Bottom line: don't SIGHUP unless you have to.  Ok?
 */
void reloadconfig(int x)
{
#ifdef WITH_DEBUG
	printf("Configuration being reloaded\n");
#endif
	config.close();
	config.loadFile(sysconfdir "/antinat.conf");
	signal (x, reloadconfig);
}

void closeup(int x)
{
#ifdef WITH_DEBUG
	printf("Closing up\n");
#endif
	logEvent(LOG_EVT_SERVERCLOSE);
}

/*
main() - you've seen that before.
*/
int
main (int argc, char * argv[])
{
	DSParams param;
	DSComponent auth(&versions);
	DSComponent filter(&filters);
	DSComponent address(&addresses);
	DSComponent logs(&loggers);
	DSComponent res(&resolvers);
	DSComponent revres(&revresolve);
	int i;
	param.setFlagSwitch('x');
	param.process(argc,argv);
#ifndef WITH_DEBUG
	if (!param.getNumericValue('x')) daemonize();
#endif
	config.loadFile(sysconfdir "/antinat.conf");
	i=address.loadcomponents(componentdir "/address", "address_type",
			"getfunction");
	if (i==0) {
		fprintf(stderr,"Could not load any address type handlers.  Antinat cannot continue.\n");
		return EXIT_FAILURE;
	}
#ifdef WITH_DEBUG
	printf("Loaded %i address components\n",i);
#endif
	i=filter.loadcomponents(componentdir "/filters", "filter_type",
			"filter");
#ifdef WITH_DEBUG
	printf("Loaded %i filter components\n",i);
#endif
	i=logs.loadcomponents(componentdir "/loggers", "logger_type",
			"log");
#ifdef WITH_DEBUG
	printf("Loaded %i logger components\n",i);
#endif
	i=auth.loadcomponents(componentdir "/versions", "version_type",
			"version_process");
	if (i==0) {
		fprintf(stderr,"Could not load any version handlers.  Antinat cannot continue.\n");
		return EXIT_FAILURE;
	}
#ifdef WITH_DEBUG
	printf("Loaded %i version components\n",i);
#endif
	i=res.loadcomponents(componentdir "/resolvers", "resolver_type",
			"resolve");
#ifdef WITH_DEBUG
	printf("Loaded %i resolver components\n",i);
#endif
	i=revres.loadcomponents(componentdir "/revres", "resolver_type",
			"resolve");
#ifdef WITH_DEBUG
	printf("Loaded %i reverse resolver components\n",i);
#endif

	signal (SIGPIPE, ignorer);
	signal (SIGHUP, reloadconfig);
	signal (SIGQUIT, closeup);
	if (!server.startServer (config.getValueInt("boot","port",1080)))
	{
		printf ("Could not initialise server.\n");
		return (EXIT_FAILURE);
	}
	logEvent(LOG_EVT_SERVERSTART);
	while (TRUE)
	{
		if (!HandleRequest ())
		{
			printf ("Couldn't handle request.\n");
		}
	}
	return EXIT_SUCCESS;
}

